/* global Promise */
(function (exports) {
  'use strict';

  var oauth3 = {};
  var logins = {};

  var core = exports.OAUTH3_CORE || require('./oauth3.core.js');

  oauth3.requests = logins;

  if ('undefined' !== typeof Promise) {
    oauth3.PromiseA = Promise;
  } else {
    console.warn("[oauth3.js] Remember to call oauth3.providePromise(Promise) with a proper Promise implementation");
  }

  oauth3.providePromise = function (PromiseA) {
    oauth3.PromiseA = PromiseA;
    if (oauth3._testPromise) {
      return oauth3._testPromise(PromiseA).then(function () {
        oauth3.PromiseA = PromiseA;
      });
    }

    oauth3.PromiseA = PromiseA;
    return PromiseA.resolve();
  };

  oauth3._recaseRequest = function (recase, req) {
    // convert JavaScript camelCase to oauth3/ruby snake_case
    if (req.data && 'object' === typeof req.data) {
      req.originalData = req.data;
      req.data = recase.snakeCopy(req.data);
    }

    return req;
  };
  oauth3._recaseResponse = function (recase, resp) {
    // convert oauth3/ruby snake_case to JavaScript camelCase
    if (resp.data && 'object' === typeof resp.data) {
      resp.originalData = resp.data;
      resp.data = recase.camelCopy(resp.data);
    }
    return resp;
  };
  oauth3._lintRequest = function (preq, opts) {
    var providerUri;
    var fresh;

    console.log('[os] request meta opts', opts);

    // check that the JWT is not expired
    // TODO check that this request applies to the aud and azp
    if (!(preq.session && preq.session.accessToken)) {
      console.log('[os] no session/accessTokenData');
      return oauth3.PromiseA.resolve(preq);
    }

    preq.headers = preq.headers || {};
    preq.headers.Authorization = 'Bearer ' + preq.session.accessToken;

    if (!preq.session._accessTokenData) {
      console.log('[os] no _accessTokenData');
      preq.session._accessTokenData = core.jwt.decode(preq.session.accessToken).payload;
    }

    if (!preq.url.match(preq.session._accessTokenData.aud)) {
      console.log("[os] doesn't match audience", preq.session._accessTokenData.aud);
      return oauth3.PromiseA.resolve(preq);
    }

    fresh = (Date.now() / 1000) >= (parseInt(preq.session._accessTokenData.exp) || 0);
    if (!fresh) {
      console.log("[os] isn't fresh", preq.session._accessTokenData.exp);
      return oauth3.PromiseA.resolve(preq);
    }

    if (!preq.session.refreshToken) {
      console.log("[os] cann't refresh", preq.session);
      return oauth3.PromiseA.resolve(preq);
    }

    opts.refreshToken = preq.session.refreshToken;
    console.log('[oauth3.js] refreshToken attempt');

    // TODO include directive?
    providerUri = preq.session.providerUri || preq.session._accessTokenData.iss;
    //opts.
    return oauth3.refreshToken(providerUri, opts).then(function (res) {
      console.log('[oauth3.js] refreshToken result:', res);

      if (!res.data.accessToken) {
        return preq;
      }

      // TODO fire session update event
      res.data.providerUri = preq.session.providerUri;
      preq.session = res.data;
      preq.headers.Authorization = 'Bearer ' + preq.session.accessToken;
      return preq;
    });
  };

  oauth3.provideRequest = function (rawRequest, opts) {
    opts = opts || {};
    var Recase = exports.Recase || require('recase');
    // TODO make insensitive to providing exceptions
    var recase = Recase.create({ exceptions: {} });

    function lintAndRequest(preq) {
      return oauth3._lintRequest(preq, opts).then(function (preq) {
        return rawRequest(preq);
      });
    }

    if (opts.rawCase) {
      oauth3.request = lintAndRequest;
      return;
    }

    // Wrap oauth3 api calls in snake_case / camelCase conversion
    oauth3.request = function (req, opts) {
      //console.log('[D] [oauth3 req.url]', req.url);
      opts = opts || {};

      if (opts.rawCase) {
        return lintAndRequest(req, opts);
      }

      req = oauth3._recaseRequest(recase, req);
      return lintAndRequest(req, opts).then(function (res) {
        return oauth3._recaseResponse(recase, res);
      });
    };

    /*
    return oauth3._testRequest(request).then(function () {
      oauth3.request = request;
    });
    */
  };

  oauth3.loginCode = function (providerUri, opts) {
    return oauth3.discover(providerUri, opts).then(function (directive) {
      var prequest = core.loginCode(directive, opts);

      console.log('[DEBUG] [core] loginCode URL', prequest);

      return oauth3.request(prequest).then(function (res) {
        // result = { uuid, expiresAt }
        console.log('[DEBUG] [core] loginCode result', res);
        return {
          otpUuid: res.data.uuid
        , otpExpires: res.data.expiresAt
        };
      });
    });
  };

  oauth3.resourceOwnerPassword = function (providerUri, username, passphrase, opts) {
    console.log('DEBUG oauth3.resourceOwnerPassword opts', opts);
    //var scope = opts.scope;
    //var appId = opts.appId;
    return oauth3.discover(providerUri, opts).then(function (directive) {
      var prequest = core.resourceOwnerPassword(directive, opts);

      console.log('[DEBUG] [core] resourceOwnerPassword URL', prequest);

      return oauth3.request({
        url: prequest.url
      , method: prequest.method
      , data: prequest.data
      });
    });
  };

  oauth3.refreshToken = function (providerUri, opts) {
    return oauth3.discover(providerUri, opts).then(function (directive) {
      var prequest = core.refreshToken(directive, opts);

      console.log('[DEBUG] [core] refreshToken URL', prequest);

      return oauth3.request({
        url: prequest.url
      , method: prequest.method
      , data: prequest.data
      });
    });
  };

  // TODO It'll be very interesting to see if we can do some browser popup stuff from the CLI
  logins._error_description = 'Not Implemented: Please override by including <script src="oauth3.browser.js"></script>';
  logins.authorizationRedirect = function (/*providerUri, opts*/) {
    throw new Error(logins._error_description);
  };
  logins.implicitGrant = function (/*providerUri, opts*/) {
    throw new Error(logins._error_description);
  };
  logins.logout = function (/*providerUri, opts*/) {
    throw new Error(logins._error_description);
  };

  oauth3.login = function (providerUri, opts) {
    console.log('##### DEBUG oauth3.login providerUri, opts');
    console.log(providerUri);
    console.log(opts);
    // Four styles of login:
    //   * background (hidden iframe)
    //   * iframe (visible iframe, needs border color and width x height params)
    //   * popup (needs width x height and positioning? params)
    //   * window (params?)

    // Two strategies
    //  * authorization_redirect (to server authorization code)
    //  * implicit_grant (default, browser-only)
    // If both are selected, implicit happens first and then the other happens in background

    var promise;

    if (opts.username || opts.password) {
      /* jshint ignore:start */
      // ingore "confusing use of !"
      if (!opts.username !== !(opts.password || opts.otp)) {
        throw new Error("you did not specify both username and password");
      }
      /* jshint ignore:end */

      var username = opts.username;
      var password = opts.password;
      delete opts.username;
      delete opts.password;

      return oauth3.resourceOwnerPassword(providerUri, username, password, opts).then(function (resp) {
        if (!resp || !resp.data) {
          var err = new Error("bad response");
          err.response = resp;
          err.data = resp && resp.data || undefined;
          return oauth3.PromiseA.reject(err);
        }
        return resp.data;
      });
    }

    // TODO support dual-strategy login
    // by default, always get implicitGrant (for client)
    // and optionally do authorizationCode (for server session)
    if ('background' === opts.type || opts.background) {
      opts.type = 'background';
      opts.background = true;
    }
    else {
      opts.type = 'popup';
      opts.popup = true;
    }
    if (opts.authorizationRedirect) {
      promise = logins.authorizationRedirect(providerUri, opts);
    }
    else {
      promise = logins.implicitGrant(providerUri, opts);
    }

    return promise;
  };

  oauth3.backgroundLogin = function (providerUri, opts) {
    opts = opts || {};
    opts.type = 'background';
    return oauth3.login(providerUri, opts);
  };

  oauth3.core = core;
  oauth3.querystringify = core.querystringify;
  oauth3.scopestringify = core.stringifyscope;
  oauth3.stringifyscope = core.stringifyscope;

  exports.OAUTH3 = oauth3.oauth3 = oauth3.OAUTH3 = oauth3;
  exports.oauth3 = exports.OAUTH3;

  if ('undefined' !== typeof module) {
    module.exports = oauth3;
  }
}('undefined' !== typeof exports ? exports : window));
